/*
 *  powerswitch.c
 *  Joris van Rantwijk
 *
 *  Powerswitch module for Linux 2.2.x - 2.5.x kernels.
 *
 *  Compile with optimizations and load as a module, like this:
 *    gcc -Wall -O2 -I/usr/src/linux/include -o powerswitch.o -c powerswitch.c
 *    insmod powerswitch.o
 */

#define __KERNEL__
#define MODULE
#define __KERNEL_SYSCALLS__
#define KBUILD_MODNAME powerswitch

/* Deal with SMP and symbol versioning */
#include <linux/config.h>
#ifdef CONFIG_SMP
#define __SMP__
#endif
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/stddef.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/unistd.h>
#include <asm/uaccess.h>

/* Make sure we have the KERNEL_VERSION symbol */
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) (((a) << 16) + ((b) << 8) + (c))
#endif

/* Use a bit of magic to choose the right pieces of code */
#if (! defined(USE_PM)) && (! defined(USE_APMDEV))
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
#error "Unsupported version of the Linux kernel"
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define USE_PM
#else
#define USE_APMDEV
#endif
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
/* Introduced somewhere during 2.4.0-test series */
#define HAVE_CALLUSERMODEHELPER
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,61)
#define OLD_CALL_USERMODEHELPER(p,a,e) call_usermodehelper(p,a,e,0)
#else
#define OLD_CALL_USERMODEHELPER call_usermodehelper
#endif
#endif

/* Support new vermagic tricks */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/vermagic.h>
MODULE_INFO(vermagic, VERMAGIC_STRING);
#endif

#ifdef USE_PM
#include <linux/pm.h>
#define INTERFACE_TYPE "pm"
#endif

#ifdef USE_APMDEV
#include <linux/apm_bios.h>
static inline _syscall3(int,ioctl,int,fd,int,cmd,long,arg);
static inline _syscall1(int,chdir,char*,filename);
#define INTERFACE_TYPE "apm_bios"
#endif

#ifdef HAVE_CALLUSERMODEHELPER
#include <linux/kmod.h>
#endif


MODULE_AUTHOR("Joris van Rantwijk");
MODULE_DESCRIPTION("Power switch shutdown support");

#ifdef EXPORT_NO_SYMBOLS
EXPORT_NO_SYMBOLS;
#endif

#define MODULE_NAME	"powerswitch: "
#define MODULE_VERSION	"1.5"
#define SHUTDOWN_PROG	"/sbin/shutdown"

static int errno;

#ifdef USE_PM
static struct pm_dev *pmdev = NULL;
#endif

#ifdef USE_APMDEV
#define APM_DEVICE	"/dev/apm_bios"
static pid_t powerswitch_pid = 0;
static int apm_fd;
#endif

static char *shutdown_argv[] = { SHUTDOWN_PROG, "-h", "now", NULL };
static char *shutdown_envp[] = { NULL };


#if defined(USE_APMDEV) || (! defined(HAVE_CALLUSERMODEHELPER))
static int do_shutdown(void *arg)
{
#ifdef USE_APMDEV
	sigemptyset(&current->blocked);
	close(apm_fd);
#endif

	set_fs(KERNEL_DS);

	open("/dev/console", O_RDWR, 0);
	dup(0);
	dup(0);

	execve(SHUTDOWN_PROG, shutdown_argv, shutdown_envp);

	printk(KERN_ERR MODULE_NAME
	       "Cannot execute " SHUTDOWN_PROG "\n");

	return 1;
}
#endif


#ifdef USE_PM
static int pmcallback(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	if (rqst == PM_SUSPEND) {

		/* Power button press !! */
		printk(KERN_NOTICE MODULE_NAME
		       "Received suspend event, initiating shutdown\n");

		/* The kapmd is trying to be clever by setting its state
		   to TASK_INTERRUPTIBLE even while it is running.
		   We know better than that! */
		set_current_state(TASK_RUNNING);

#ifdef HAVE_CALLUSERMODEHELPER
		/* Start the shutdown program in usermode */
		if ( OLD_CALL_USERMODEHELPER(SHUTDOWN_PROG,
		       shutdown_argv, shutdown_envp) < 0 ) {
			printk(KERN_ERR MODULE_NAME
			       "Cannot execute " SHUTDOWN_PROG "\n");
		}
#else
		/* The kapmd calls exit_files(current) at startup;
		   this seems to make it impossible to use execve() without
		   oopsing. We play a very dirty trick to temporarily undo
		   the effects of exit_files(). */
		{ struct files_struct *oldfiles;
		  if ((oldfiles = current->files) == NULL)
			current->files = init_task.files;

		  /* Spawn a kernel thread to execute the shutdown program */
		  kernel_thread(do_shutdown, NULL, 0);
		  current->files = oldfiles;
		}
#endif

		return -EBUSY;

	} else {

		/* Ignore other requests */
		return 0;

	}
}
#endif


#ifdef USE_APMDEV
static int mainloop(void *arg)
{
	struct task_struct *tsk = current;
	apm_event_t apm_event;
	ssize_t n;
	int i;

	/* Fixup task structure */
	tsk->session = 1;
	tsk->pgrp = 1;
	tsk->tty = NULL;
	tsk->tty_old_pgrp = 0;
	strcpy(tsk->comm, "powerswitch");

	/* Drop memory pages */
	exit_mm(tsk);

	/* Block all signals except SIGKILL */
	sigfillset(&tsk->blocked);
	sigdelset(&tsk->blocked, SIGKILL);

	/* This seems necessary, although I don't understand why. */
	set_fs(KERNEL_DS);

	/* Close open file descriptors */
	for (i = 0; i < NR_OPEN; i++)
		close(i);

	chdir("/");

	/* Open APM device */
	apm_fd = open(APM_DEVICE, O_RDONLY, 0);
	if (apm_fd < 0) {
		printk(KERN_ERR MODULE_NAME
		       "Cannot open " APM_DEVICE " (err %d)\n", errno);
		MOD_DEC_USE_COUNT;
		return 1;
	}

	/* Wait for APM event notifications */
	while (1) {

		n = read(apm_fd, (char *) &apm_event, sizeof(apm_event_t));

		if (n != sizeof(apm_event_t)) {
			if (n == -1 && errno != ERESTARTSYS)
				printk(KERN_ERR MODULE_NAME
				  "Error reading from APM device (err %d)\n",
				  errno);
			else if (n >= 0)
				printk(KERN_ERR MODULE_NAME
				  "Incomplete event data from APM device\n");
			printk(KERN_NOTICE MODULE_NAME
			       "Terminating APM monitor thread\n");
			MOD_DEC_USE_COUNT;
			return 1;
		}

		if (apm_event == APM_SYS_STANDBY ||
		    apm_event == APM_USER_STANDBY) {
			/* Acknowledge standby events */
			ioctl(apm_fd, APM_IOC_STANDBY, 0);
		}

		if (apm_event == APM_SYS_SUSPEND ||
		    apm_event == APM_USER_SUSPEND) {
			/* Power button press !! */
			break;
		}

		/* Ignore other events */

	}

	/* Block any further SIGKILL signals */
	sigaddset(&tsk->blocked, SIGKILL);

	if (signal_pending(tsk)) {
		/* Pending KILL signal, terminate the thread */
		printk(KERN_NOTICE MODULE_NAME
		       "Terminating APM monitor thread\n");
		MOD_DEC_USE_COUNT;
		return 1;
	}

	printk(KERN_NOTICE MODULE_NAME
	       "Received suspend event, initiating shutdown\n");

	/* Spawn new thread to execute /sbin/shutdown */
	i = kernel_thread(do_shutdown, NULL, 0);

	/* Sleep forever to keep /dev/apm_bios open */
	tsk->state = TASK_INTERRUPTIBLE;
	schedule_timeout(MAX_SCHEDULE_TIMEOUT);

	MOD_DEC_USE_COUNT;
	return 0;
}
#endif


int powerswitch_init(void)
{
	printk(KERN_INFO MODULE_NAME
	       "Module initializing, version " MODULE_VERSION
	       " (" INTERFACE_TYPE "), Joris van Rantwijk\n");

#ifdef USE_PM
	pmdev = pm_register(PM_UNKNOWN_DEV, PM_SYS_UNKNOWN, pmcallback);

	if (pmdev == NULL) {
		printk(KERN_ERR MODULE_NAME
		       "Cannot register powermanagement callback handler\n");
		return -EBUSY;
	}
#endif

#ifdef USE_APMDEV
	/* Starting a "kernel thread" from a user process context is
	   probably not a very clever thing to do. But I cannot think
	   of any better way. */

	powerswitch_pid = kernel_thread(mainloop, NULL, 0);
	if (powerswitch_pid < 0) {
		printk(KERN_ERR MODULE_NAME
		       "Cannot create APM monitor thread (err %d)\n",
		       powerswitch_pid);
		return powerswitch_pid;
	}

	/* Bad things happen if the module is removed while the thread
	   is still running. (Trust me, I tried it a few times) */

	MOD_INC_USE_COUNT;

	printk(KERN_INFO MODULE_NAME
	       "APM monitor thread created (pid = %d)\n", powerswitch_pid);
#endif

	return 0;
}


void powerswitch_cleanup(void)
{
	printk(KERN_INFO MODULE_NAME "Removing module\n");

#ifdef USE_PM
	pm_unregister(pmdev);
#endif
}

module_init(powerswitch_init);
module_exit(powerswitch_cleanup);

/* end */
